#!/usr/bin/env python
"""Examples of python APIs."""

import numpy as np
import spglib


def _show_symmetry(symmetry):
    for i in range(symmetry["rotations"].shape[0]):
        print("  --------------- %4d ---------------" % (i + 1))
        rot = symmetry["rotations"][i]
        trans = symmetry["translations"][i]
        print("  rotation:")
        for x in rot:
            print("     [%2d %2d %2d]" % (x[0], x[1], x[2]))
        print("  translation:")
        print("     (%8.5f %8.5f %8.5f)" % (trans[0], trans[1], trans[2]))


def _show_lattice(lattice):
    print("Basis vectors:")
    for vec, axis in zip(lattice, ("a", "b", "c")):
        print(
            "%s %10.5f %10.5f %10.5f"
            % (
                tuple(
                    axis,
                )
                + tuple(vec)
            ),
        )


def _show_cell(lattice, positions, numbers):
    _show_lattice(lattice)
    print("Atomic points:")
    for p, s in zip(positions, numbers):
        print("%2d %10.5f %10.5f %10.5f" % ((s,) + tuple(p)))


def main():
    """Run examples."""
    silicon = (
        [(4, 0, 0), (0, 4, 0), (0, 0, 4)],
        [
            (0, 0, 0),
            (0, 0.5, 0.5),
            (0.5, 0, 0.5),
            (0.5, 0.5, 0),
            (0.25, 0.25, 0.25),
            (0.25, 0.75, 0.75),
            (0.75, 0.25, 0.75),
            (0.75, 0.75, 0.25),
        ],
        [
            14,
        ]
        * 8,
    )

    silicon_dist = (
        [(4.01, 0, 0), (0, 4, 0), (0, 0, 3.99)],
        [
            (0.001, 0, 0),
            (0, 0.5, 0.5),
            (0.5, 0, 0.5),
            (0.5, 0.5, 0),
            (0.25, 0.25, 0.251),
            (0.25, 0.75, 0.75),
            (0.75, 0.25, 0.75),
            (0.75, 0.75, 0.25),
        ],
        [
            14,
        ]
        * 8,
    )

    silicon_prim = (
        [(0, 2, 2), (2, 0, 2), (2, 2, 0)],
        [(0, 0, 0), (0.25, 0.25, 0.25)],
        [14, 14],
    )

    rutile = (
        [(4, 0, 0), (0, 4, 0), (0, 0, 3)],
        [
            (0, 0, 0),
            (0.5, 0.5, 0.5),
            (0.3, 0.3, 0.0),
            (0.7, 0.7, 0.0),
            (0.2, 0.8, 0.5),
            (0.8, 0.2, 0.5),
        ],
        [14, 14, 8, 8, 8, 8],
    )

    rutile_dist = (
        [(3.97, 0, 0), (0, 4.03, 0), (0, 0, 3.0)],
        [
            (0, 0, 0),
            (0.5001, 0.5, 0.5),
            (0.3, 0.3, 0.0),
            (0.7, 0.7, 0.002),
            (0.2, 0.8, 0.5),
            (0.8, 0.2, 0.5),
        ],
        [14, 14, 8, 8, 8, 8],
    )

    a = 3.07
    c = 3.52
    MgB2 = (
        [(a, 0, 0), (-a / 2, a / 2 * np.sqrt(3), 0), (0, 0, c)],
        [(0, 0, 0), (1.0 / 3, 2.0 / 3, 0.5), (2.0 / 3, 1.0 / 3, 0.5)],
        [12, 5, 5],
    )

    a = [3.0, 0.0, 0.0]
    b = [-3.66666667, 3.68178701, 0.0]
    c = [-0.66666667, -1.3429469, 1.32364995]
    niggli_lattice = np.array([a, b, c])

    print("[get_spacegroup]")
    print("  Spacegroup of Silicon is %s." % spglib.get_spacegroup(silicon))
    print("")
    print("[get_spacegroup]")
    print("  Spacegroup of Rutile is %s." % spglib.get_spacegroup(rutile))
    print("")
    print("[get_spacegroup]")
    print("  Spacegroup of MgB2 is %s." % spglib.get_spacegroup(MgB2))
    print("")
    print("[get_symmetry]")
    print("  Symmetry operations of Rutile unitcell are:")
    print("")
    symmetry = spglib.get_symmetry(rutile)
    _show_symmetry(symmetry)
    print("")
    print("[get_symmetry]")
    print("  Symmetry operations of MgB2 are:")
    print("")
    symmetry = spglib.get_symmetry(MgB2)
    _show_symmetry(symmetry)
    print("")
    print("[get_pointgroup]")
    print(
        "  Pointgroup of MgB2 is %s." % spglib.get_pointgroup(symmetry["rotations"])[0],
    )
    print("")

    dataset = spglib.get_symmetry_dataset(rutile)
    print("[get_symmetry_dataset] ['international']")
    print(
        "  Spacegroup of Rutile is %s (%d)." % (dataset.international, dataset.number),
    )
    print("")
    print("[get_symmetry_dataset] ['pointgroup']")
    print("  Pointgroup of Rutile is %s." % (dataset.pointgroup))
    print("")
    print("[get_symmetry_dataset] ['hall']")
    print(
        "  Hall symbol of Rutile is %s (%d)." % (dataset.hall, dataset.hall_number),
    )
    print("")
    print("[get_symmetry_dataset] ['wyckoffs']")
    print("  Wyckoff letters of Rutile are: ", dataset.wyckoffs)
    print("")
    print("[get_symmetry_dataset] ['equivalent_atoms']")
    print("  Mapping to equivalent atoms of Rutile are: ")
    for i, x in enumerate(dataset.equivalent_atoms):
        print("  %d -> %d" % (i + 1, x + 1))
    print("")
    print("[get_symmetry_dataset] ['rotations'], ['translations']")
    print("  Symmetry operations of Rutile unitcell are:")
    for i, (rot, trans) in enumerate(
        zip(dataset.rotations, dataset.translations),
    ):
        print("  --------------- %4d ---------------" % (i + 1))
        print("  rotation:")
        for x in rot:
            print("     [%2d %2d %2d]" % (x[0], x[1], x[2]))
        print("  translation:")
        print("     (%8.5f %8.5f %8.5f)" % (trans[0], trans[1], trans[2]))
    print("")

    print("[refine_cell]")
    print(" Refine distorted rutile structure")
    lattice, positions, numbers = spglib.refine_cell(rutile_dist, symprec=1e-1)
    _show_cell(lattice, positions, numbers)
    print("")

    print("[find_primitive]")
    print(" Fine primitive distorted silicon structure")
    lattice, positions, numbers = spglib.find_primitive(silicon_dist, symprec=1e-1)
    _show_cell(lattice, positions, numbers)
    print("")

    print("[standardize_cell]")
    print(" Standardize distorted rutile structure:")
    print(" (to_primitive=0 and no_idealize=0)")
    lattice, positions, numbers = spglib.standardize_cell(
        rutile_dist,
        to_primitive=0,
        no_idealize=0,
        symprec=1e-1,
    )
    _show_cell(lattice, positions, numbers)
    print("")

    print("[standardize_cell]")
    print(" Standardize distorted rutile structure:")
    print(" (to_primitive=0 and no_idealize=1)")
    lattice, positions, numbers = spglib.standardize_cell(
        rutile_dist,
        to_primitive=0,
        no_idealize=1,
        symprec=1e-1,
    )
    _show_cell(lattice, positions, numbers)
    print("")

    print("[standardize_cell]")
    print(" Standardize distorted silicon structure:")
    print(" (to_primitive=1 and no_idealize=0)")
    lattice, positions, numbers = spglib.standardize_cell(
        silicon_dist,
        to_primitive=1,
        no_idealize=0,
        symprec=1e-1,
    )
    _show_cell(lattice, positions, numbers)
    print("")

    print("[standardize_cell]")
    print(" Standardize distorted silicon structure:")
    print(" (to_primitive=1 and no_idealize=1)")
    lattice, positions, numbers = spglib.standardize_cell(
        silicon_dist,
        to_primitive=1,
        no_idealize=1,
        symprec=1e-1,
    )
    _show_cell(lattice, positions, numbers)
    print("")

    symmetry = spglib.get_symmetry(silicon)
    print("[get_symmetry]")
    print("  Number of symmetry operations of silicon conventional")
    print("  unit cell is %d (192)." % len(symmetry["rotations"]))
    _show_symmetry(symmetry)
    print("")

    symmetry = spglib.get_symmetry_from_database(525)
    print("[get_symmetry_from_database]")
    print("  Number of symmetry operations of silicon conventional")
    print("  unit cell is %d (192)." % len(symmetry["rotations"]))
    _show_symmetry(symmetry)
    print("")

    reduced_lattice = spglib.niggli_reduce(niggli_lattice)
    print("[niggli_reduce]")
    print("  Original lattice")
    _show_lattice(niggli_lattice)
    print("  Reduced lattice")
    _show_lattice(reduced_lattice)
    print("")

    mapping, grid = spglib.get_ir_reciprocal_mesh(
        [11, 11, 11],
        silicon_prim,
        is_shift=[0, 0, 0],
    )
    num_ir_kpt = len(np.unique(mapping))
    print("[get_ir_reciprocal_mesh]")
    print("  Number of irreducible k-points of primitive silicon with")
    print("  11x11x11 Monkhorst-Pack mesh is %d (56)." % num_ir_kpt)
    print("")

    mapping, grid = spglib.get_ir_reciprocal_mesh([8, 8, 8], rutile, is_shift=[1, 1, 1])
    num_ir_kpt = len(np.unique(mapping))
    print("[get_ir_reciprocal_mesh]")
    print("  Number of irreducible k-points of Rutile with")
    print("  8x8x8 Monkhorst-Pack mesh is %d (40)." % num_ir_kpt)
    print("")

    mapping, grid = spglib.get_ir_reciprocal_mesh([9, 9, 8], MgB2, is_shift=[0, 0, 1])
    num_ir_kpt = len(np.unique(mapping))
    print("[get_ir_reciprocal_mesh]")
    print("  Number of irreducible k-points of MgB2 with")
    print("  9x9x8 Monkhorst-Pack mesh is %s (48)." % num_ir_kpt)
    print("")


if __name__ == "__main__":
    main()



import os
import unittest

import numpy as np
from spglib import (
    SpglibDataset,
    get_symmetry,
    get_symmetry_dataset,
    get_symmetry_from_database,
)
from vasp import read_vasp

data_dir = os.path.dirname(os.path.abspath(__file__))

dirnames = (
    "cubic",
    "hexagonal",
    "monoclinic",
    "orthorhombic",
    "tetragonal",
    "triclinic",
    "trigonal",
    "distorted",
    "virtual_structure",
)


class TestChangeOfBasis(unittest.TestCase):
    def setUp(self):
        self._filenames = []
        self._datasets: list[SpglibDataset] = []
        self._cells = []
        self._symprecs = []
        for d in dirnames:
            dirname = os.path.join(data_dir, "data", d)
            filenames = os.listdir(dirname)
            self._filenames += [os.path.join(dirname, fname) for fname in filenames]

        for fname in self._filenames:
            cell = read_vasp(fname)
            if "distorted" in fname:
                symprec = 1e-1
            else:
                symprec = 1e-5
            self._datasets.append(get_symmetry_dataset(cell, symprec=symprec))
            self._cells.append(cell)
            self._symprecs.append(symprec)

    def tearDown(self):
        pass

    def test_change_of_basis(self):
        for fname, dataset, cell, symprec in zip(
            self._filenames,
            self._datasets,
            self._cells,
            self._symprecs,
        ):
            std_pos = dataset.std_positions
            tmat = dataset.transformation_matrix
            orig_shift = dataset.origin_shift
            lat = np.dot(cell[0].T, np.linalg.inv(tmat))
            pos = np.dot(cell[1], tmat.T) + orig_shift
            for p in pos:
                diff = std_pos - p
                diff -= np.rint(diff)
                diff = np.dot(diff, lat.T)
                delta = np.sqrt((diff**2).sum(axis=1))
                indices = np.where(delta < symprec)[0]
                self.assertEqual(len(indices), 1, msg=("multi: %s %s" % (p, indices)))

    def test_std_symmetry(self):
        for fname, dataset, cell, symprec in zip(
            self._filenames,
            self._datasets,
            self._cells,
            self._symprecs,
        ):
            symmetry = get_symmetry_from_database(dataset.hall_number)
            std_pos = dataset.std_positions

            # for r, t in zip(symmetry['rotations'], symmetry['translations']):
            #     for rp in (np.dot(std_pos, r.T) + t):
            #         diff = std_pos - rp
            #         diff -= np.rint(diff)
            #         num_match = len(np.where(abs(diff).sum(axis=1) < 1e-3)[0])
            #         self.assertEqual(num_match, 1, msg="%s" % fname)

            # Equivalent above by numpy hack
            # 15 sec on macOS 2.3 GHz Intel Core i5 (4times faster than above)
            rot = symmetry["rotations"]
            trans = symmetry["translations"]
            # (n_sym, 3, n_atom)
            rot_pos = np.dot(rot, std_pos.T) + trans[:, :, None]
            for p in std_pos:
                diff = rot_pos - p[None, :, None]
                diff -= np.rint(diff)
                num_match = (abs(diff).sum(axis=1) < 1e-3).sum(axis=1)
                self.assertTrue((num_match == 1).all(), msg="%s" % fname)

    def test_std_rotation(self):
        for fname, dataset, cell, symprec in zip(
            self._filenames,
            self._datasets,
            self._cells,
            self._symprecs,
        ):
            std_lat = dataset.std_lattice
            tmat = dataset.transformation_matrix
            lat = np.dot(cell[0].T, np.linalg.inv(tmat))
            lat_rot = np.dot(dataset.std_rotation_matrix, lat)
            np.testing.assert_allclose(
                std_lat,
                lat_rot.T,
                atol=symprec,
                err_msg="%s" % fname,
            )

    def _test_get_symmetry(self):
        """Caution!: This test must be executed with spglib compiled with -DSPGTEST."""
        for fname, dataset, cell, symprec in zip(
            self._filenames,
            self._datasets,
            self._cells,
            self._symprecs,
        ):
            cell_spin = cell + (
                [
                    1,
                ]
                * len(cell[2]),
            )
            symmetry = get_symmetry(cell_spin, symprec=symprec)

            self.assertEqual(
                len(dataset.rotations),
                len(symmetry["rotations"]),
                msg=("%s" % fname),
            )

            for r, t in zip(dataset.rotations, dataset.translations):
                found = False
                for r_, t_ in zip(symmetry["rotations"], symmetry["translations"]):
                    if (r == r_).all():
                        diff = t - t_
                        diff -= np.rint(diff)
                        if (abs(diff) < symprec).all():
                            found = True
                            break
                self.assertTrue(found, msg="%s" % fname)


if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(TestChangeOfBasis)
    unittest.TextTestRunner(verbosity=2).run(suite)
    # unittest.main()



import unittest

import numpy as np
from spglib import get_magnetic_symmetry, get_symmetry


class TestGetSymmetry(unittest.TestCase):
    def setUp(self):
        lattice = [[4, 0, 0], [0, 4, 0], [0, 0, 4]]
        positions = [[0, 0, 0], [0.5, 0.5, 0.5]]
        numbers = [1, 1]
        magmoms = [0, 0]
        self._cell = (lattice, positions, numbers, magmoms)

    def tearDown(self):
        pass

    def test_deprecation_warning(self):
        self._cell[3][0] = 1
        self._cell[3][1] = 1
        with self.assertWarns(DeprecationWarning):
            sym = get_symmetry(self._cell)
        self.assertEqual(96, len(sym["rotations"]))
        np.testing.assert_equal(sym["equivalent_atoms"], [0, 0])

    def test_get_symmetry_ferro(self):
        self._cell[3][0] = 1
        self._cell[3][1] = 1
        sym = get_magnetic_symmetry(self._cell)
        self.assertEqual(96, len(sym["rotations"]))
        np.testing.assert_equal(sym["equivalent_atoms"], [0, 0])

    def test_get_symmetry_anti_ferro(self):
        self._cell[3][0] = 1
        self._cell[3][1] = -1
        sym = get_magnetic_symmetry(self._cell)
        self.assertEqual(96, len(sym["rotations"]))
        np.testing.assert_equal(sym["equivalent_atoms"], [0, 0])

    def test_get_symmetry_broken_magmoms(self):
        self._cell[3][0] = 1
        self._cell[3][1] = 2
        sym = get_magnetic_symmetry(self._cell)
        self.assertEqual(48, len(sym["rotations"]))
        np.testing.assert_equal(sym["equivalent_atoms"], [0, 1])


if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(TestGetSymmetry)
    unittest.TextTestRunner(verbosity=2).run(suite)
    # unittest.main()



from spglib import __version__ as version

project = "Spglib"
copyright = "2009, Atsushi Togo"

extensions = [
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinxcontrib.bibtex",
    "myst_parser",
    "autodoc2",
    "sphinx.ext.doctest",
]

exclude_patterns = [
    "README.md",
    "_build",
]
source_suffix = {
    ".md": "markdown",
    ".rst": "restructuredtext",
}

# -----------------------------------------------------------------------------
# MyST
# -----------------------------------------------------------------------------

myst_enable_extensions = [
    "amsmath",
    "dollarmath",
    "html_admonition",
    "html_image",
    "replacements",
    "smartquotes",
    "tasklist",
    "colon_fence",
]

myst_dmath_double_inline = True
myst_heading_anchors = 3

# -----------------------------------------------------------------------------
# BibTeX
# -----------------------------------------------------------------------------

bibtex_bibfiles = ["references.bib"]
bibtex_default_style = "unsrt"

# -----------------------------------------------------------------------------
# Sphinx-book-theme
# -----------------------------------------------------------------------------

html_theme = "sphinx_book_theme"
html_title = f"Spglib v{version}"
html_theme_options = {
    # https://sphinx-book-theme.readthedocs.io/en/latest/reference.html
    "repository_url": "https://github.com/spglib/spglib",
    "use_repository_button": True,
    "navigation_with_keys": True,
    "globaltoc_includehidden": "true",
    "show_toc_level": 3,
}
html_static_path = ["_static"]

# -----------------------------------------------------------------------------
# Autodoc2
# -----------------------------------------------------------------------------
autodoc2_output_dir = "api/python-api"
autodoc2_render_plugin = "myst"
autodoc2_docstring_parser_regexes = [
    (r".*", "rst"),
]
autodoc2_annotations = False
autodoc2_packages = [
    {
        "path": "../python/spglib/spglib.py",
        "module": "spglib",
    },
]
autodoc2_hidden_objects = ["dunder", "private", "inherited"]
autodoc2_hidden_regexes = [
    "spglib.get_pointgroup",
    # Layer group
    "spglib.get_layergroup",
    "spglib.get_symmetry_layerdataset",
    # Kpoints
    "spglib.get_grid_point_from_address",
    "spglib.get_stabilized_reciprocal_mesh",
    "spglib.get_grid_points_by_rotations",
    "spglib.get_BZ_grid_points_by_rotations",
    "spglib.relocate_BZ_grid_address",
]

# -----------------------------------------------------------------------------
# linkcheck
# -----------------------------------------------------------------------------

linkcheck_ignore = [
    # Ignore closed-access links
    r"https://www.sciencedirect.com/.*",
    r"https://www.jstor.org/.*",
    # Cannot check for 403 error only
    r"https://doi.org/10.1002/qua.20747",
    r"https://github.com/spglib/spglib/pull/.*",
]
linkcheck_allowed_redirects = {
    r"https://doi.org/.*": r".*",
    r"https://github.com/spglib/spglib/archive/.*": r".*",
    r"https://spglib.readthedocs.io/.*": r"https://spglib.readthedocs.io/.*",
    r"https://docs.rs/crate/spglib/": r"https://docs.rs/crate/spglib/.*",
    r"https://github.com/chrisjsewell/sphinx-autodoc2": r"https://github.com/sphinx-extensions2/sphinx-autodoc2",
}
linkcheck_anchors_ignore_for_url = [
    r"https://github.com",
]



#!/usr/bin/env python

import sys
import numpy as np

lattice_symbols = {
    'P': [[0, 0, 0]],
    'A': [[0, 0, 0], [0, 12, 12]],
    'B': [[0, 0, 0], [12, 0, 12]],
    'C': [[0, 0, 0], [12, 12, 0]],
    'I': [[0, 0, 0], [12, 12, 12]],
    'R': [[0, 0, 0]],
    'H': [[0, 0, 0], [16, 8, 8], [8, 16, 16]],
    'F': [[0, 0, 0], [0, 12, 12], [12, 0, 12], [12, 12, 0]],
    'p': [[0, 0, 0]],
    'c': [[0, 0, 0], [12, 12, 0]]
}


def position2operation(p):
    r = np.zeros((3, 3), dtype=int)
    xyz = 'xyz'

    for i in range(3):
        for j, c in enumerate(xyz):
            pos = p[i].find(c)
            # if exist xyz
            if pos > -1:
                if pos == 0:
                    r[i, j] = 1
                if pos > 0:
                    if p[i][pos-1] == '+':  # +y, +z
                        r[i, j] = 1
                    elif p[i][pos-1] == '-':  # -x, -y, -z
                        r[i, j] = -1
                    elif p[i][pos-1].isdigit():  # 2x
                        if pos > 1:
                            if p[i][pos-2] == '-':
                                r[i, j] = - int(p[i][pos-1])
                            else:
                                r[i, j] = int(p[i][pos-1])
                        else:
                            r[i, j] = int(p[i][pos-1])
                    else:
                        print("error")
                        sys.exit(1)

                # print p, p[i], r[i,j]

    t = np.zeros(3, dtype=int)

    # translation multiplied by 24
    try:
        table = str.maketrans({'x': '', 'y': '', 'z': '', '+': '', '-': ''})
    except AttributeError:
        pass
    for i in range(3):
        try:
            trans = p[i].translate(None, 'xyz+-')
        except TypeError:
            trans = p[i].translate(table)
        pos = trans.find('/')
        if pos > -1:
            frac = trans.split('/')
            t[i] = int(frac[0]) * (24 / int(frac[1]))

    return r, t


def encode_rotation(r):
    enc = []
    for i in range(3):
        enc.append((r[i][0] + 2) * 9 + (r[i][1] + 1) * 3 + r[i][2] + 1)
    return enc[0] * 45**2 + enc[1] * 45 + enc[2]


def decode_rotation(c):
    r = np.zeros((3, 3), dtype=int)
    dec = [c // 45**2, (c % 45**2) // 45, c % 45]
    for i in range(3):
        r[i][0] = dec[i] // 9 - 2
        r[i][1] = (dec[i] % 9) // 3 - 1
        r[i][2] = dec[i] % 3 - 1
    return r


def decode_trans(c):
    return c // 576, (c % 576) // 24, (c % 24)


def run_test(len_wyckoffs, len_site, o_flat):
    sum_elem = 1
    for i in range(1, 531):
        sum_site = 0
        for j in range(len_wyckoffs[i], len_wyckoffs[i+1]):
            print(len_site[j+1] - len_site[j]),

            sum_site += len_site[j+1] - len_site[j]

            for x in o_flat[len_site[j]:len_site[j+1]]:
                t = decode_trans(x // 3**9)
                r = decode_rotation(x % 3**9)
                print(r, t)

        print("  %d" % sum_site)
        sum_elem += sum_site
    return sum_elem


def damp_array_wyckoffs(len_wyckoff):
    print("static const int position_wyckoff[] =")
    text = "  { %4d," % 0
    for i, x in enumerate(len_wyckoffs[1:]):
        if i % 10 == 0:
            print(text)
            text = "   "
        text += " %4d," % x
    print(text + " };")


def damp_array_numsites(len_site):
    text = "  { %5d," % 0
    for i, x in enumerate(len_site[1:]):
        if i % 10 == 0:
            print(text)
            text = "   "
        text += " %5d," % x
    print(text + " };")


def damp_array_positions(orbits):
    text = "  { %10d," % 0
    for i, x in enumerate(orbits[1:]):
        if i % 5 == 0:
            print(text)
            text = "   "
        text += " %10d," % x
    print(text + " };")


def damp_array_positions_short(orbits, len_site):
    count = 0
    print("static const int coordinates_first[] =")
    text = "  { %9d," % 0
    for i, x in enumerate(len_site[1:-1]):
        if count % 5 == 0:
            print(text)
            text = "   "
        text += " %9d," % orbits[x]
        count += 1
    print(text + " };")

    count = 0
    print("static const int num_sitesym[] =")
    text = "  { %3d," % 0
    for i, x in enumerate(len_site[1:-1]):
        if count % 10 == 0:
            print(text)
            text = "   "
        text += " %3d," % (len_site[i+2] - len_site[i+1])
        count += 1
    print(text + " };")


def damp_array_site_symmetries(site_symmetries):
    print("static const char site_symmetries[][7] =")
    text = "  { %8s," % "\"      \""
    for i, x in enumerate(site_symmetries):
        if i % 6 == 0:
            print(text)
            text = "   "
        text += " %8s," % x
    print(text + " };")


def read_wyckoff_csv(filename):
    with open(filename) as wyckoff_file:
        return parse_wyckoff_csv(wyckoff_file)


def parse_wyckoff_csv(wyckoff_file):
    """Parse Wyckoff.csv

    There are 530 data sets. For one example:

    9:C 1 2 1:::::::
    ::4:c:1:(x,y,z):(-x,y,-z)::
    ::2:b:2:(0,y,1/2):::
    ::2:a:2:(0,y,0):::

    """

    rowdata = []
    points = []
    hP_nums = [433, 436, 444, 450, 452, 458, 460]
    for i, line in enumerate(wyckoff_file):
        if line.strip() == 'end of data':
            break
        rowdata.append(line.strip().split(':'))

        # 2:P -1  ::::::: <-- store line number if first element is number
        if rowdata[-1][0].isdigit():
            points.append(i)
    points.append(i)

    wyckoff = []
    for i in range(len(points) - 1):  # 0 to 529
        symbol = rowdata[points[i]][1]  # e.g. "C 1 2 1"
        if i + 1 in hP_nums:
            symbol = symbol.replace('R', 'H', 1)
        wyckoff.append({'symbol': symbol.strip()})

    # When the number of positions is larger than 4,
    # the positions are written in the next line.
    # So those positions are connected.
    for i in range(len(points) - 1):
        count = 0
        wyckoff[i]['wyckoff'] = []
        for j in range(points[i] + 1, points[i + 1]):
            # Hook if the third element is a number (multiplicity), e.g.,
            #
            # 232:P 2/b 2/m 2/b:::::::  <- ignored
            # ::8:r:1:(x,y,z):(-x,y,-z):(x,-y+1/2,-z):(-x,-y+1/2,z)
            # :::::(-x,-y,-z):(x,-y,z):(-x,y+1/2,z):(x,y+1/2,-z)  <- ignored
            # ::4:q:..m:(x,0,z):(-x,0,-z):(x,1/2,-z):(-x,1/2,z)
            # ::4:p:..2:(0,y,1/2):(0,-y+1/2,1/2):(0,-y,1/2):(0,y+1/2,1/2)
            # ::4:o:..2:(1/2,y,0):(1/2,-y+1/2,0):(1/2,-y,0):(1/2,y+1/2,0)
            # ...
            if rowdata[j][2].isdigit():
                pos = []
                w = {'letter': rowdata[j][3].strip(),
                     'multiplicity': int(rowdata[j][2]),
                     'site_symmetry': rowdata[j][4].strip(),
                     'positions': pos}
                wyckoff[i]['wyckoff'].append(w)

                for k in range(4):
                    if rowdata[j][k + 5]:  # check if '(x,y,z)' or ''
                        count += 1
                        pos.append(rowdata[j][k + 5])
            else:
                for k in range(4):
                    if rowdata[j][k + 5]:
                        count += 1
                        pos.append(rowdata[j][k + 5])

        # assertion
        for w in wyckoff[i]['wyckoff']:
            n_pos = len(w['positions'])
            n_pos *= len(lattice_symbols[wyckoff[i]['symbol'][0]])
            assert n_pos == w['multiplicity']

    return wyckoff


def get_wyckoff_positions(wyckoff):
    positions = []
    for i, w in enumerate(wyckoff):
        # print "%d:%s" % (i+1, w['symbol'])
        positions_hall = []
        for p in w['wyckoff']:
            # print "  %d %s %s" % (p['multiplicity'],
            #                        p['letter'],
            #                        p['site_symmetry']),
            # print
            # for o in p['positions']:
            #     print o,
            # print

            positions_site = []
            for pure_trans in lattice_symbols[w['symbol'][0]]:
                for o in p['positions']:
                    r, t = position2operation(o.strip('()').split(','))
                    t += np.array(pure_trans)
                    t = t % 24
                    positions_site.append((r, t))
            positions_hall.append(positions_site)
        positions.append(positions_hall)
    return positions


def get_site_symmetries(wyckoff):
    """List up site symmetries

    The data structure is as follows:

        wyckoff[0]['wyckoff'][0]['site_symmetry']

    Note
    ----
    Maximum length of string is 6.

    """

    ssyms = []
    for w in wyckoff:
        ssyms += ["\"%-6s\"" % w_s['site_symmetry'] for w_s in w['wyckoff']]

    damp_array_site_symmetries(ssyms)


def encode_wyckoff_positions(positions):
    maxval = 0
    maxmulti = 0
    maxr = None
    maxt = None
    number = 0
    pos_encoded = []
    numerators = np.zeros(24, dtype=int)

    for i, positions_hall in enumerate(positions):
        pos_encoded_hall = []
        for positions_site in positions_hall:
            pos_encoded_site = []
            for r, t in positions_site:
                # for j in range(3):
                #     numerators[t[j]] += 1

                t_encode = t[0] * 576 + t[1] * 24 + t[2]
                r_encode = encode_rotation(r)
                total = t_encode * 45**3 + r_encode

                pos_encoded_site.append(total)

                if (r - decode_rotation(r_encode)).any():
                    print("Error in r %d" % (i+1))
                    sys.exit(1)
                # else:
                #     if (abs(decode_rotation(r_encode)) > 1).any():
                #         print i+1
                #         print decode_rotation(r_encode)
                #         print t
                #         print
                if (t - decode_trans(t_encode)).any():
                    print("Error in t")
                    sys.exit(1)

                if total > maxval:
                    maxval = total
                    maxr = r.copy()
                    maxt = t.copy()
                    number = i+1

            pos_encoded_hall.append(pos_encoded_site)
        pos_encoded.append(pos_encoded_hall)

    # for i, n in enumerate(numerators):
    #     print i, n

    # print "MAX: %d" % maxmulti
    # print "MAX: %d (%d)" % (maxval,number)
    # print maxr
    # print maxt

    return pos_encoded


def get_data_arrays(encoded_positions):
    len_wyckoffs = [0]
    len_site = [0]
    sum_wyckoffs = 1
    sum_site = 1
    encpos_flat = [0]
    for h in encoded_positions:
        len_wyckoffs.append(sum_wyckoffs)
        sum_wyckoffs += len(h)
        for w in h:
            len_site.append(sum_site)
            sum_site += len(w)
            encpos_flat += w

    len_wyckoffs.append(sum_wyckoffs)
    len_site.append(sum_site)

    return len_wyckoffs, len_site, encpos_flat


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.set_defaults(is_site=False,
                        is_position=False,
                        is_wyckoff=False,
                        is_site_short=False,
                        is_site_symbol=False,
                        is_wyckoff_short=False)

    parser.add_argument("--site", dest="is_site",
                        action="store_true",
                        help="Index of site symmetries of each space group")
    parser.add_argument("--site-symbol", dest="is_site_symbol",
                        action="store_true",
                        help="Site symmetry symbols")
    parser.add_argument("--position", dest="is_position",
                        action="store_true",
                        help=("Index of positions of each site symmetries and "
                              "space group"))
    parser.add_argument("--wyckoff", dest="is_wyckoff",
                        action="store_true",
                        help="Encoded Wyckoff position operators")
    parser.add_argument("--wyckoff_short", dest="is_wyckoff_short",
                        action="store_true",
                        help="Encoded Wyckoff position operators")
    parser.add_argument("filename", nargs='?',
                        help="Filename of csv file")
    args = parser.parse_args()
    wyckoff = read_wyckoff_csv(args.filename)
    operations = get_wyckoff_positions(wyckoff)
    encoded_positions = encode_wyckoff_positions(operations)
    len_wyckoffs, len_site, encpos_flat = get_data_arrays(encoded_positions)

    # Test
    # sum_elem = run_test(len_wyckoffs, len_site, o_flat)
    # print "%d = %d" % (len(o_flat), sum_elem)

    # Damp full arrays
    if args.is_position:
        damp_array_wyckoffs(len_wyckoffs)
    if args.is_site:
        damp_array_numsites(len_site)
    if args.is_wyckoff:
        damp_array_positions(encpos_flat)
    if args.is_wyckoff_short:
        damp_array_positions_short(encpos_flat, len_site)
    if args.is_site_symbol:
        get_site_symmetries(wyckoff)



import unittest

import numpy as np
from spglib import get_magnetic_symmetry, get_symmetry


class TestGetOperationsWithSiteTensors(unittest.TestCase):
    def setUp(self):
        lattice = [[0, 2, 2], [2, 0, 2], [2, 2, 0]]
        positions = [[0, 0, 0]]
        numbers = [1]
        magmoms = [[0, 0, 1]]
        self._cell_Ni = (lattice, positions, numbers, magmoms)

        # CrCl2, adapted from Example 1 in Tutorial-MAXMAGN:
        # http://webbdcrista1.ehu.es/cryst/tutorials/Tutorial-MAXMAGN.pdf
        lattice = [
            [6.8257, 0, 0],
            [0, 6.2139, 0],
            [0, 0, 3.4947],
        ]
        positions = [
            # Cr (2a), site symmetry: < -x,-y,-z; -x,-y,z >
            [0, 0, 0],
            [0.5, 0.5, 0.5],
            # Cl (4g)
            [0.3586, 0.2893, 0],
            [-0.3586, -0.2893, 0],
            [-0.3586 + 0.5, 0.2893 + 0.5, 0.5],
            [0.3586 + 0.5, -0.2893 + 0.5, 0.5],
        ]
        numbers = [1, 1, 2, 2, 2, 2]
        self._cell_CrCl2 = (lattice, positions, numbers)

    def tearDown(self):
        pass

    def test_get_symmetry_non_collinear(self):
        sym = get_magnetic_symmetry(self._cell_Ni, with_time_reversal=False)
        self.assertEqual(8, len(sym["rotations"]))
        np.testing.assert_equal(sym["equivalent_atoms"], [0])

        # type-III magnetic space group
        sym2 = get_magnetic_symmetry(self._cell_Ni)
        self.assertEqual(16, len(sym2["rotations"]))
        self.assertEqual(16, len(sym2["translations"]))
        self.assertEqual(16, len(sym2["time_reversals"]))

    def test_get_symmetry_vectors(self):
        # Space group without magnetic moments
        # Pnnm (58), "-P 2 2n" (hall_number=275)
        # Generators: -x,-y,-z; -x,-y,z; -x+1/2,y+1/2,-z+1/2
        sym = get_symmetry(self._cell_CrCl2)
        self.assertEqual(8, len(sym["rotations"]))

        # Type-I magnetic space group (MSG)
        magmoms1 = [
            [0, 0, 1],
            [0, 0, -1],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
        ]
        sym1 = get_magnetic_symmetry(self._cell_CrCl2 + (magmoms1,))
        self.assertTrue(np.allclose(sym1["time_reversals"], False))
        self.assertTrue(np.allclose(sym1["rotations"], sym["rotations"]))
        self.assertTrue(np.allclose(sym1["translations"], sym["translations"]))

        # Type-II MSG
        # 58.394, -P 2 2n 1'
        magmoms2 = [
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
        ]
        sym2 = get_magnetic_symmetry(self._cell_CrCl2 + (magmoms2,))
        self.assertEqual(len(sym2["rotations"]), 16)
        self.assertEqual(np.sum(sym2["time_reversals"]), 8)

        sym2_gray = get_magnetic_symmetry(
            self._cell_CrCl2 + (magmoms2,),
            with_time_reversal=False,
        )
        self.assertTrue(np.allclose(sym2_gray["rotations"], sym["rotations"]))
        self.assertTrue(np.allclose(sym2_gray["translations"], sym["translations"]))

        # Type-III MSG
        # 58.397: -P 2 2n'
        # Generators: -x,-y,-z; -x,-y,z; -x+1/2,y+1/2,-z+1/2'
        magmoms3 = [
            [0, 0, 1],
            [0, 0, 1],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
        ]
        sym3 = get_magnetic_symmetry(self._cell_CrCl2 + (magmoms3,))
        self.assertTrue(np.allclose(sym3["rotations"], sym["rotations"]))
        self.assertTrue(np.allclose(sym3["translations"], sym["translations"]))
        self.assertTrue(np.sum(sym3["time_reversals"]), 4)

        # Type-IV MSG
        # https://github.com/spglib/spglib/issues/150
        # -P 2 2 -> -P 2 2 1c' (47.254)
        lat = [[1, 0, 0], [0, 2, 0], [0, 0, 3]]
        pos = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.5]]
        num = [0, 0]
        magmom = [[0.0, 0.0, 1.0], [0.0, 0.0, -1.0]]
        sym_gray = get_symmetry((lat, pos, num))
        sym4 = get_magnetic_symmetry((lat, pos, num, magmom))
        self.assertTrue(np.allclose(sym_gray["rotations"], sym4["rotations"]))
        self.assertTrue(np.allclose(sym_gray["translations"], sym4["translations"]))
        self.assertTrue(sym4["time_reversals"].shape[0], 16)
        self.assertTrue(np.sum(sym4["time_reversals"]), 8)
        self.assertTrue(np.allclose(sym4["primitive_lattice"], lat))

    def test_with_distorted_magmom(self):
        magmoms = [
            [0, 0, 1.001],
            [0, 0, 0.999],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
        ]
        sym = get_magnetic_symmetry(self._cell_CrCl2 + (magmoms,), mag_symprec=1e-2)
        expect = [0, 0, 2, 2, 2, 2]
        assert np.allclose(sym["equivalent_atoms"], expect)


if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(
        TestGetOperationsWithSiteTensors,
    )
    unittest.TextTestRunner(verbosity=2).run(suite)
    # unittest.main()



from __future__ import annotations

from pathlib import Path
from typing import Callable

import pytest
from spglib import get_symmetry_dataset


@pytest.mark.benchmark(group="space-group")
def test_get_symmetry_dataset(
    benchmark,
    all_filenames: list[Path],
    read_vasp: Callable,
):
    """Benchmarking get_symmetry_dataset on all structures under test/data."""
    # Load all cells beforehand
    cells = [read_vasp(fname) for fname in all_filenames]
    print(f"Benchmark get_symmetry_dataset on {len(cells)} structures")

    def _get_symmetry_dataset_for_cells():
        for cell in cells:
            _ = get_symmetry_dataset(cell, symprec=1e-5)

    benchmark.pedantic(_get_symmetry_dataset_for_cells, rounds=4)



from __future__ import annotations

import sys

import numpy as np
from numpy.typing import NDArray

from operation import MagneticOperation, traverse

sys.path.append("../")
from hall2operations import lattice_symbols, rotation_matrices, translations


class MagneticHallSymbol:
    """
    Parser for magnetic Hall symbol. See A1.4.2.3 in ITB (2010).

    Extended Backus-Naur form (EBNF) for Magnetic Hall symbols
    ----------------------------------------------------------
    <Hall symbol>    := <L> <N>+ <V>?
    <L>              := "-"? <lattice symbol>
    <lattice symbol> := [PABCIRHF]  # Table A1.4.2.2
    <N>              := <nfold> <A>? <T>? "'"?  # "'" represents time reversal
    <nfold>          := "-"? ("1" | "2" | "3" | "4" | "6")
    For SG,  <A>     := [xyz] | "'" | '"' | "=" | "*"  # Table A.1.4.2.4, Table A.1.4.2.5, Table A.1.4.2.6
    For MSG, <A>     := [xyz] | "^" | '"' | "=" | "*"  # The original axis symbol "'" is substituted with "^"
    <T>              := [abcnuvwd] | [1-6]  # Table A.1.4.2.3
    <V>              := "(" [0-11] [0-11] [0-11] ")"
    """
    def __init__(self, hall_symbol: str):
        self._hall_symbol = hall_symbol
        self._parse(magnetic=True)
        self._coset = traverse(self.generators)

    @property
    def hall_symbol(self) -> str:
        return self._hall_symbol

    @property
    def inversion_at_origin(self) -> bool:
        return self._inversion_at_origin

    @property
    def lattice_symbol(self) -> str:
        return self._lattice_symbol

    @property
    def generators(self):
        return self._generators

    @property
    def coset(self):
        """
        Coset representative of magnetic space group over its translational subgroup
        """
        return self._coset

    def _parse(self, magnetic: bool = True):
        tokens = self.hall_symbol.split(" ")

        assert len(tokens) >= 1
        self._parse_L(tokens[0])

        self._Ns = []
        rotation_count = 0

        self._V = np.zeros(3)
        prev_nfold = None
        prev_axis = None
        for i in range(1, len(tokens)):
            if tokens[i][0] == "(":
                # tokens[i:] = ["(" + vx, vy, vz + ")"]
                self._V = self._parse_V(tokens[i:])
                break
            else:
                # We need to remember the location of "N" symbols and preceding "N" symbol
                # because its default axis depends on them!
                rotation, translation, nfold, axis, time_reversal = self._parse_N(
                    tokens[i], rotation_count, prev_nfold, prev_axis, magnetic
                )
                self._Ns.append((rotation, translation, time_reversal))
                prev_nfold = nfold
                prev_axis = axis
                rotation_count += 1

        # change basis by (I, v)
        # (R, tau) -> (R, tau + v - Rv)
        self._generators = []
        for t in lattice_symbols[self.lattice_symbol]:  # type: ignore
            if not np.allclose(t, 0):
                ops = MagneticOperation.from_linear_translation_time_reversal(translation=np.array(t))
                self._generators.append(ops)

        if self.inversion_at_origin:
            r, t = self.change_of_basis(-np.eye(3), np.zeros(3))
            ops = MagneticOperation.from_linear_translation_time_reversal(linear=r, translation=t)
            self._generators.append(ops)

        for rotation, translation, time_reversal in self._Ns:
            r, t = self.change_of_basis(rotation, translation)
            ops = MagneticOperation.from_linear_translation_time_reversal(linear=r, translation=t, time_reversal=time_reversal)
            self._generators.append(ops)

    def _parse_L(self, token):
        if token[0] == "-":
            self._inversion_at_origin = True
            self._lattice_symbol = token[1:]
        else:
            self._inversion_at_origin = False
            self._lattice_symbol = token

    @staticmethod
    def _parse_N(
        token, count: int, prev_nfold: str | None, prev_axis: str | None, magnetic: bool = False
    ):
        pos = 0

        # improper
        improper = False
        if token[pos] == "-":
            pos += 1
            improper = True

        # nfold
        nfold = token[pos]
        pos += 1

        # change axis symbol for MHall
        if magnetic:
            prime_axis_symbol = "^"
        else:
            prime_axis_symbol = "'"

        # axis
        axis = ""
        if pos <= len(token) - 1:
            if token[pos] == prime_axis_symbol:
                axis = "p"
                pos += 1
            elif token[pos] in ['"', '=']:
                # "
                axis = "pp"
                pos += 1

        if pos <= len(token) - 1:
            if token[pos] in ["x", "y", "z", "*"]:
                axis = token[pos]
                pos += 1
        if (axis in ["p", "pp"]) and (prev_axis in ["x", "y", "z"]):
            # See Table A1.4.2.5
            axis = axis + prev_axis

        # No need to specify axis for nfold=1
        if nfold == "1":
            axis = "z"

        # Default axes. See A1.4.2.3.1
        if axis in ["", "p", "pp"]:
            if count == 0:
                # axis direction of c
                axis += "z"
            elif count == 1:
                if prev_nfold in ["2", "4"]:
                    # axis direction of a
                    axis += "x"
                elif prev_nfold in ["3", "6"]:
                    # axis direction of a-b
                    axis = "pz"
                else:
                    raise ValueError("unreachable!")
            elif count == 2:
                if nfold == "3":
                    # axis direction of a+b+c
                    axis += "*"
                else:
                    raise ValueError("unreachable!")
            else:
                raise ValueError("unreachable!")

        rotation = np.array(rotation_matrices[nfold + axis])
        if improper:
            rotation = -rotation

        # translation
        translation = np.zeros(3)
        while pos <= len(token) - 1:
            # translations are applied additively
            if token[pos] in "123456":
                # always along z-axis!
                translation = [0, 0, int(token[pos]) / abs(int(nfold))]
            elif token[pos] in "abcnuvwd":
                translation += translations[token[pos]]
            else:
                break
            pos += 1

        # For magnetic hall symbols
        time_reversal = False
        if magnetic and (pos < len(token)) and (token[pos] == "'"):
            pos += 1
            time_reversal = True

        assert pos == len(token)
        return rotation, translation, nfold, axis, time_reversal

    @staticmethod
    def _parse_V(tokens):
        assert len(tokens) == 3
        assert tokens[0][0] == "("
        vx = int(tokens[0][1:])
        vy = int(tokens[1])
        vz = int(tokens[2][:-1])
        assert tokens[2][-1] == ")"

        translation = np.array([vx / 12, vy / 12, vz / 12])

        return translation

    def change_of_basis(self, rotation: NDArray, translation: NDArray):
        new_translation = translation + self._V - np.dot(rotation, self._V)
        return rotation, new_translation



from __future__ import annotations

import sys
import os
import subprocess
from tempfile import TemporaryDirectory
from string import Template
import json
from itertools import product

from joblib import Parallel, delayed
import numpy as np
from numpy.typing import NDArray
import click

from operation import MagneticOperation
from make_msgtype_db import get_type_of_msg
from magnetic_hall import MagneticHallSymbol
from transform import Transformation, get_standard_hall_number
from load import get_msg_numbers, get_spg_table, get_msg_table

sys.path.append('..')
from hall2operations import encode_symmetry


MAX_ENCODED = (3 ** 9) * (12 ** 3)
MAX_DENOMINATOR = 4


def enumerate_linears(max_element=1):
    # Enumerate integer matrix with determinant=1
    linears = []
    for nums in product(range(-max_element, max_element + 1), repeat=9):
        arr = np.array(nums).reshape(3, 3)
        if np.isclose(np.linalg.det(arr), 1):
            linears.append(arr.astype(int))
    return linears

def enumerate_origin_shifts():
    shifts = []
    for shift in product([0, 1 / 4, 1 / 3, 1 / 2, 2 / 3, 3 / 4], repeat=3):
        shifts.append(np.array(shift))
    return shifts


def get_family_space_group(coset: list[MagneticOperation]) -> list[MagneticOperation]:
    fsg = []
    for g in coset:
        fsg.append(MagneticOperation.from_linear_translation_time_reversal(
            linear=g.linear,
            translation=g.translation,
            time_reversal=False,
        ))
    return fsg


def get_maximal_space_subgroup(coset: list[MagneticOperation]) -> list[MagneticOperation]:
    xsg = []
    for g in coset:
        if g.time_reversal:
            continue
        xsg.append(MagneticOperation.from_linear_translation_time_reversal(
            linear=g.linear,
            translation=g.translation,
            time_reversal=False,
        ))
    return xsg


def dump_operation(g: MagneticOperation) -> str:
    if g.time_reversal:
        raise ValueError("Only accept ordinary operation.")

    ret = "[ "
    for i in range(3):
        ret += f"[{g._linear_tuple[i][0]}, {g._linear_tuple[i][1]}, {g._linear_tuple[i][2]}, {g._translation_tuple[i]}], "
    ret += "[0, 0, 0, 1] ]"

    return ret


def get_conjugator_type3(coset: list[MagneticOperation], linears, shifts) -> list[NDArray]:
    fsg = set(get_family_space_group(coset))
    pg = set([ops._linear_tuple for ops in fsg])

    found = [set(coset)]
    conjugators = []
    for linear in linears:
        t_no_shift = Transformation(linear=linear)
        coset2_no_shift = t_no_shift.transform_coset(coset)
        fsg2_no_shift = set(get_family_space_group(coset2_no_shift))
        pg2_no_shift = set([ops._linear_tuple for ops in fsg2_no_shift])

        # At least, point groups should be equal
        if pg2_no_shift != pg:
            continue

        for origin_shift in shifts:
            t = Transformation(linear=linear, origin_shift=origin_shift)
            coset2 = t.transform_coset(coset)
            fsg2 = set(get_family_space_group(coset2))

            if fsg2 != fsg:
                continue
            coset2 = set(coset2)
            if coset2 in found:
                continue
            found.append(coset2)
            conjugators.append(t._augmented_matrix.tolist())
            break

    return conjugators


def get_conjugator_type4(coset: list[MagneticOperation], linears, shifts) -> list[NDArray]:
    """
    Return conjugators s.t.
        (P, p)^-1 * coset * (P, p)
    is conjugate with `coset`.
    """
    xsg = set(get_maximal_space_subgroup(coset))
    pg = set([ops._linear_tuple for ops in xsg])

    found = [set(coset)]
    conjugators = []
    for linear in linears:
        t_no_shift = Transformation(linear=linear)
        coset2_no_shift = t_no_shift.transform_coset(coset)
        xsg2_no_shift = set(get_maximal_space_subgroup(coset2_no_shift))
        pg2_no_shift = set([ops._linear_tuple for ops in xsg2_no_shift])

        # At least, point groups should be equal
        if pg2_no_shift != pg:
            continue

        for origin_shift in shifts:
            t = Transformation(linear=linear, origin_shift=origin_shift)
            coset2 = t.transform_coset(coset)
            xsg2 = set(get_maximal_space_subgroup(coset2))

            if xsg2 != xsg:
                continue
            coset2 = set(coset2)
            if coset2 in found:
                continue
            found.append(coset2)
            conjugators.append(t._augmented_matrix.tolist())
            break

    return conjugators


def get_conjugator(hall_number, spg_table, msg_table, mapping, linears, shifts):
    data = {}
    number = spg_table[hall_number]['number']
    choice = spg_table[hall_number]['choice']
    hall_number_std = get_standard_hall_number(number)
    # Transformation from ITA standard settings to the other settings
    trns = Transformation.to_standard(hall_number, choice, number).inverse()

    for bns_number, mhall_symbol in msg_table[hall_number_std].items():
        mhs = MagneticHallSymbol(mhall_symbol)
        coset = trns.transform_coset(mhs.coset)
        msg_type = get_type_of_msg(mhall_symbol)
        uni_number = mapping[bns_number]

        if msg_type == 3 and number >= 16:
            # Triclinic and monoclinic seem to have no alternative setting for type-III
            ret = get_conjugator_type3(coset, linears, shifts)
        elif msg_type == 4:
            ret = get_conjugator_type4(coset, linears, shifts)
        else:
            ret = []

        data[(hall_number, uni_number)] = ret
        # print(f"hall_number={hall_number}, uni_number={uni_number} ({msg_type}): {ret}")

    return data


def dump_cpp(raw_all_datum):
    all_datum = {}
    for data in raw_all_datum:
        conjugators = data['conjugators']
        all_datum[(data['hall_number'], data['uni_number'])] = conjugators

        # Check if symmetry encoding can be used
        for g in conjugators:
            assert np.max(np.abs(g)) <= 1

    mapping = {uni_number: [] for uni_number in range(1, 1651 + 1)}
    for (hall_number, uni_number), conjugators in all_datum.items():
        # sorted by uni_number->hall_number
        mapping[uni_number].append((hall_number, conjugators))

    MAX_HALL_NUMBERS = 18
    MAX_NUM_SETTINGS = 6

    contents = []
    contents.append(f"static const int alternative_transformations[][{MAX_HALL_NUMBERS}][{MAX_NUM_SETTINGS + 1}] = {{")
    contents.append("    { { 0 } }, /* dummy */")
    for uni_number in range(1, 1651 + 1):
        hall_conjugators = mapping[uni_number]
        assert len(hall_conjugators) <= MAX_HALL_NUMBERS
        contents.append("    {")
        for hall_number, conjugators in hall_conjugators:
            line = "        { "
            for conj in conjugators:
                conj = np.array(conj)
                linear = np.around(conj[:-1, :-1]).astype(int)
                assert np.isclose(np.linalg.det(linear), 1)  # Sanity check
                origin_shift = np.remainder(conj[:-1, -1], 1)
                encoded = encode_symmetry(linear, origin_shift)
                assert 0 <= encoded < MAX_ENCODED
                line += f"{encoded}, "
            line += f"0 }}, /* UNI={uni_number}, Hall={hall_number} */"
            contents.append(line)
        contents.append("    },")

    contents.append("};")

    for line in contents:
        print(line)


@click.command()
@click.option('--output', default="conjugators.json")
@click.option('--dump', is_flag=True, default=False)
def main(output, dump):
    if dump:
        with open(output, 'r') as f:
            all_datum = json.load(f)
        dump_cpp(all_datum)
        return

    msg_table = get_msg_table()
    spg_table = get_spg_table()

    msg_numbers = get_msg_numbers()
    mapping = {}
    for _, bns_number, _, uni_number in msg_numbers:
        mapping[bns_number] = uni_number

    linears = enumerate_linears()
    shifts = enumerate_origin_shifts()

    ret = Parallel(n_jobs=-1, verbose=11)(delayed(get_conjugator)(hall_number, spg_table, msg_table, mapping, linears, shifts) for hall_number in range(1, 530 + 1))

    all_datum = []
    for r in ret:
        for (hall_number, uni_number), val in r.items():
            all_datum.append({
                'hall_number': hall_number,
                'uni_number': uni_number,
                'conjugators': val,
            })

    with open(output, 'w') as f:
        json.dump(all_datum, f, indent=2)

    return all_datum


def debug():
    linears = enumerate_linears()
    shifts = enumerate_origin_shifts()

    # Testing
    # hall_symbol ="P 2ac 2ab'"
    # hall_symbol = "P 2ac' 2ab -1'"  # 61.435
    # hall_symbol = "-P 2 2'"  # 47.252
    hall_symbol = "P 2' 2'"  # 16.3

    mhs = MagneticHallSymbol(hall_symbol)
    ret = get_conjugator_type3(mhs.coset, linears, shifts)
    print(ret)

    # hall_symbol = "P 2ac 2ab 1c'"
    # hall_symbol = "-C 2yc 1c'"
    # mhs = MagneticHallSymbol(hall_symbol)
    # ret = get_conjugator_type4(mhs.coset, linears)
    # print(ret)


if __name__ == '__main__':
    # debug()
    main()



import pytest
from spglib.spglib import _expand_cell


def test_expand_cell():
    with pytest.raises(TypeError, match="cell has to be a tuple of 3 or 4 elements."):
        _expand_cell(([], [], [], [], []))



#!/usr/bin/env python

import numpy as np
import hall2operations as h2o
import make_Wyckoff_db as wdb

characters = "abcdefghijklmnopqrstuvwxyz"

def get_Hall_operations( hall_file, hall_number ):
    hall_symbols = h2o.read_spg_csv( hall_file  )
    return h2o.HallSymbol( hall_symbols[ hall_number - 1 ][0] ).get_full_operations()

def get_wyckoff_position_operators( wyckoff_file, hall_number ):
    wyckoff = wdb.read_wyckoff_csv( wyckoff_file )
    operations = wdb.get_wyckoff_positions( wyckoff )
    return operations[ hall_number - 1 ]

def multiply_operations( spg_op, pos_op ):
    r, t = spg_op
    o_r, o_t = pos_op
    return ( np.dot( r, o_r ), np.dot( o_t, r.T ) + t )

def show_site_symmetries( rotations, translations, operators ):
    for i, o_site in enumerate( operators ):
        for j, o in enumerate( o_site ):
            o_r = o[0]
            o_t = o[1] / 24.0
            print "--------- %s, %d ----------" % (characters[len(operators) - i - 1], j+1)
            for r, t in zip( rotations, translations ):
                mo = multiply_operations( ( r, t ), ( o_r, o_t ) )
                diff = mo[1] - o_t
                if ( mo[0] - o_r == 0 ).all() and ( abs( diff - diff.round() ) < 0.01 ).all():
                    print r, t


if __name__ == '__main__':
    from optparse import OptionParser
    parser = OptionParser()
    parser.set_defaults( is_site = False )
    parser.add_option("--site", dest="is_site",
                      action="store_true",
                      help="Show site symmetries")
    (options, args) = parser.parse_args()

    rotations, translations = get_Hall_operations( args[0], int( args[2] ) )
    operators = get_wyckoff_position_operators( args[1], int( args[2] ) )


    if options.is_site:
        show_site_symmetries( rotations, translations, operators )



"""Example to use get_symmetry_dataset with Wurtzite structure input (P6_3mc)."""

from spglib import get_symmetry_dataset

lattice = [[3.111, 0, 0], [-1.5555, 2.6942050311733885, 0], [0, 0, 4.988]]
position = [
    [1.0 / 3, 2.0 / 3, 0.0],
    [2.0 / 3, 1.0 / 3, 0.5],
    [1.0 / 3, 2.0 / 3, 0.6181],
    [2.0 / 3, 1.0 / 3, 0.1181],
]
types = [1, 1, 2, 2]
symprec = 1e-5

cell = (lattice, position, types)
dataset = get_symmetry_dataset(cell, symprec=symprec)

print(f"International symbol: {dataset.international} ({dataset.number})")
print(f"Hall symbol: {dataset.hall}")
print("Wyckoff letters: ", end="")
print(" ".join([f"{w}" for w in dataset.wyckoffs]))
print("Equivalent atoms:")
for i, equiv_atom in enumerate(dataset.equivalent_atoms):
    print(f"{i} -> {equiv_atom}")
print("Space group operations:")
for i, (r, t) in enumerate(zip(dataset.rotations, dataset.translations)):
    print(f"--- {i + 1} ---")
    for vec in r:
        print(f"{vec[0]:2d} {vec[1]:2d} {vec[2]:2d}")
    print(f"{t[0]:.5f} {t[1]:.5f} {t[2]:.5f}")



"""Pytest configuration."""

from __future__ import annotations

from pathlib import Path
from typing import Callable

import numpy as np
import pytest

_dirnames = (
    "cubic",
    "hexagonal",
    "monoclinic",
    "orthorhombic",
    "tetragonal",
    "triclinic",
    "trigonal",
    "distorted",
    "virtual_structure",
)


@pytest.fixture(scope="session")
def root_data_dir() -> Path:
    _data_dir = Path(__file__).parent.absolute() / "data"
    return _data_dir


@pytest.fixture(scope="session")
def dirnames(root_data_dir) -> list[Path]:
    """Return test directory names."""
    return [root_data_dir / d for d in _dirnames]


@pytest.fixture(scope="session")
def all_filenames(root_data_dir) -> list[Path]:
    """Return all filenames in test directories."""
    all_filenames = []
    for d in (root_data_dir / _d for _d in _dirnames):
        all_filenames += [d / fname for fname in d.iterdir()]
    return all_filenames


@pytest.fixture(scope="session")
def read_vasp() -> Callable:
    """Return function to read POSCAR."""

    def _read_vasp(filename):
        with open(filename) as f:
            lines = f.readlines()
            return _get_cell(lines)

    return _read_vasp


# def read_vasp_from_strings(strings):
#     return _get_cell(StringIO(strings).readlines())


def _get_cell(lines):
    line1 = [x for x in lines[0].split()]
    if _is_exist_symbols(line1):
        symbols = line1
    else:
        symbols = None

    scale = float(lines[1])

    lattice = []
    for i in range(2, 5):
        lattice.append([float(x) for x in lines[i].split()[:3]])
    lattice = np.array(lattice) * scale

    try:
        num_atoms = np.array([int(x) for x in lines[5].split()])
        line_at = 6
    except ValueError:
        symbols = [x for x in lines[5].split()]
        num_atoms = np.array([int(x) for x in lines[6].split()])
        line_at = 7

    numbers = _expand_symbols(num_atoms, symbols)

    if lines[line_at][0].lower() == "s":
        line_at += 1

    is_cartesian = False
    if lines[line_at][0].lower() == "c" or lines[line_at][0].lower() == "k":
        is_cartesian = True

    line_at += 1

    positions = []
    for i in range(line_at, line_at + num_atoms.sum()):
        positions.append([float(x) for x in lines[i].split()[:3]])

    if is_cartesian:
        positions = np.dot(positions, np.linalg.inv(lattice))

    return (lattice, positions, numbers)


def _expand_symbols(num_atoms, symbols=None):
    expanded_symbols = []
    is_symbols = True
    if symbols is None:
        is_symbols = False
    else:
        if len(symbols) != len(num_atoms):
            is_symbols = False
        else:
            for s in symbols:
                if s not in symbol_map:
                    is_symbols = False
                    break

    if is_symbols:
        for s, num in zip(symbols, num_atoms):
            expanded_symbols += [
                symbol_map[s],
            ] * num
    else:
        for i, num in enumerate(num_atoms):
            expanded_symbols += [
                i + 1,
            ] * num

    return expanded_symbols


def _is_exist_symbols(symbols):
    for s in symbols:
        if s not in symbol_map:
            return False
    return True


symbol_map = {
    "H": 1,
    "He": 2,
    "Li": 3,
    "Be": 4,
    "B": 5,
    "C": 6,
    "N": 7,
    "O": 8,
    "F": 9,
    "Ne": 10,
    "Na": 11,
    "Mg": 12,
    "Al": 13,
    "Si": 14,
    "P": 15,
    "S": 16,
    "Cl": 17,
    "Ar": 18,
    "K": 19,
    "Ca": 20,
    "Sc": 21,
    "Ti": 22,
    "V": 23,
    "Cr": 24,
    "Mn": 25,
    "Fe": 26,
    "Co": 27,
    "Ni": 28,
    "Cu": 29,
    "Zn": 30,
    "Ga": 31,
    "Ge": 32,
    "As": 33,
    "Se": 34,
    "Br": 35,
    "Kr": 36,
    "Rb": 37,
    "Sr": 38,
    "Y": 39,
    "Zr": 40,
    "Nb": 41,
    "Mo": 42,
    "Tc": 43,
    "Ru": 44,
    "Rh": 45,
    "Pd": 46,
    "Ag": 47,
    "Cd": 48,
    "In": 49,
    "Sn": 50,
    "Sb": 51,
    "Te": 52,
    "I": 53,
    "Xe": 54,
    "Cs": 55,
    "Ba": 56,
    "La": 57,
    "Ce": 58,
    "Pr": 59,
    "Nd": 60,
    "Pm": 61,
    "Sm": 62,
    "Eu": 63,
    "Gd": 64,
    "Tb": 65,
    "Dy": 66,
    "Ho": 67,
    "Er": 68,
    "Tm": 69,
    "Yb": 70,
    "Lu": 71,
    "Hf": 72,
    "Ta": 73,
    "W": 74,
    "Re": 75,
    "Os": 76,
    "Ir": 77,
    "Pt": 78,
    "Au": 79,
    "Hg": 80,
    "Tl": 81,
    "Pb": 82,
    "Bi": 83,
    "Po": 84,
    "At": 85,
    "Rn": 86,
    "Fr": 87,
    "Ra": 88,
    "Ac": 89,
    "Th": 90,
    "Pa": 91,
    "U": 92,
    "Np": 93,
    "Pu": 94,
    "Am": 95,
    "Cm": 96,
    "Bk": 97,
    "Cf": 98,
    "Es": 99,
    "Fm": 100,
    "Md": 101,
    "No": 102,
    "Lr": 103,
    "Rf": 104,
    "Db": 105,
    "Sg": 106,
    "Bh": 107,
    "Hs": 108,
    "Mt": 109,
    "Ds": 110,
    "Rg": 111,
    "Cn": 112,
    "Nh": 113,
    "Fl": 114,
    "Mc": 115,
    "Lv": 116,
    "Ts": 117,
    "Og": 118,
}



"""Test of spacegroup_type_from_symmetry."""

from __future__ import annotations

import os
from pathlib import Path
from typing import Callable

import pytest
from spglib import (
    get_spacegroup_type,
    get_spacegroup_type_from_symmetry,
    get_symmetry_dataset,
)

data_dir = os.path.dirname(os.path.abspath(__file__))


@pytest.mark.parametrize("lattice_None", [True, False])
def test_spacegroup_type_from_symmetry(
    all_filenames: list[Path],
    read_vasp: Callable,
    lattice_None: bool,
):
    """Test spacegroup_type_from_symmetry."""
    for fname in all_filenames:
        cell = read_vasp(fname)
        if lattice_None:
            lattice = None
        else:
            lattice = cell[0]
        if "distorted" in str(fname):
            dataset = get_symmetry_dataset(cell, symprec=1e-1)
            spgtype = get_spacegroup_type_from_symmetry(
                dataset.rotations,
                dataset.translations,
                lattice=lattice,
                symprec=1e-1,
            )
            if spgtype.number != dataset.number:
                print("%d != %d in %s" % (spgtype.number, dataset.number, fname))
                ref_cell = (
                    dataset.std_lattice,
                    dataset.std_positions,
                    dataset.std_types,
                )
                dataset = get_symmetry_dataset(ref_cell, symprec=1e-5)
                spgtype = get_spacegroup_type_from_symmetry(
                    dataset.rotations,
                    dataset.translations,
                    lattice=lattice,
                    symprec=1e-5,
                )
                print(
                    "Using refined cell: %d, %d in %s"
                    % (spgtype.number, dataset.number, fname),
                )
        else:
            dataset = get_symmetry_dataset(cell, symprec=1e-5)
            spgtype = get_spacegroup_type_from_symmetry(
                dataset.rotations,
                dataset.translations,
                lattice=lattice,
                symprec=1e-5,
            )

        assert spgtype.number == dataset.number, "%d != %d in %s" % (
            spgtype.number,
            dataset.number,
            fname,
        )
        spgtype_ref = get_spacegroup_type(dataset.hall_number)
        assert spgtype == spgtype_ref



from __future__ import annotations

from typing import List

import numpy as np
from numpy.typing import NDArray

from operation import MagneticOperation, remainder1_symmetry_operation


# ref: ITA Table1.5.1.1
transformation_matrix_monoclinic = {
    "a_to_b": np.array([[0, 1, 0], [0, 0, 1], [1, 0, 0]]),  # (a, b, c) -> (c, a, b)
    "-a_to_b": np.array([[0, -1, 0], [1, 0, 0], [0, 0, 1]]),  # (a, b, c) -> (b, -a, c)
    "-b_to_b": np.array([[0, 0, 1], [0, -1, 0], [1, 0, 0]]),  # (a, b, c) -> (c, -b, a)
    "c_to_b": np.array([[0, 0, 1], [1, 0, 0], [0, 1, 0]]),  # (a, b, c) -> (b, c, a)
    "-c_to_b": np.array([[1, 0, 0], [0, 0, 1], [0, -1, 0]]),  # (a, b, c) -> (a, -c, b)
    "b2_to_b1": np.array([[0, 0, -1], [0, 1, 0], [1, 0, -1]]),  # P -> P, A -> C
    "b3_to_b1": np.array([[-1, 0, 1], [0, 1, 0], [-1, 0, 0]]),  # P -> P, I -> C
}

transformation_matrix_orthorhombic = {
    "abc": np.eye(3),
    "ba-c": np.array([[0, 1, 0], [1, 0, 0], [0, 0, -1]]),
    "cab": np.array([[0, 0, 1], [1, 0, 0], [0, 1, 0]]),
    "-cba": np.array([[0, 0, -1], [0, 1, 0], [1, 0, 0]]),
    "bca": np.array([[0, 1, 0], [0, 0, 1], [1, 0, 0]]),
    "a-cb": np.array([[1, 0, 0], [0, 0, -1], [0, 1, 0]]),
}


class Transformation:
    """
    Transformation of coordinate systems

    Consider change of a coordinate system from lattice basis (a_org, b_org, c_org) and origin O to (a_new, b_new, c_new) and O'.
    The "linear part" P relates lattice basises as:
    lattice basis is changed as,
        (a_new, b_new, c_new) = (a_org, b_org, c_org) P
    We choose original basis for the origin shift,
        p = (p1_org, p2_org, p3_org)^T s.t. p1_org * a_org + p2_org * b_arg + p3_org * c_arg = O' - O
    Then, fractional coordinates are changed as:
        (x_new, y_new, z_new)^T = (P, p)^{-1} (x_org, y_org, z_org)^T
    Symmetry operation is transformed as:
        (W, w) -> (P, p)^{-1} (W, w) (P, p) = (P^{-1}WP, P^{-1}(W * p + w - p))
    """

    ndim = 3

    def __init__(self, linear: NDArray, origin_shift=None):
        self._linear = linear  # linear part
        self._origin_shift = origin_shift  # shift vector
        if origin_shift is None:
            self._origin_shift = np.zeros(self.ndim)

        self._augmented_matrix = np.zeros((self.ndim + 1, self.ndim + 1))
        self._augmented_matrix[:self.ndim, :self.ndim] = self.linear
        self._augmented_matrix[:self.ndim, -1] = self.origin_shift
        self._augmented_matrix[-1, -1] = 1

        self._invaug = np.linalg.inv(self.augmented_matrix)

    @property
    def linear(self):
        return self._linear

    @property
    def origin_shift(self):
        return self._origin_shift

    @property
    def augmented_matrix(self):
        return self._augmented_matrix

    def transform_operation(
        self, ops: MagneticOperation
    ) -> MagneticOperation:
        """
        transform lattice vector A maps to AP, where A is column-wise.
        (W, w) -> (P, p)^{-1} (W, w) (P, p) = (P^{-1}WP, P^{-1}(W * p + w - p))
        """
        new_aug = np.dot(self._invaug, np.dot(ops.augmented_matrix, self.augmented_matrix))
        return MagneticOperation(new_aug, ops.time_reversal)

    def transform_coset(
        self, coset: List[MagneticOperation]
    ) -> List[MagneticOperation]:
        transformed = []
        for ops in coset:
            remainded = remainder1_symmetry_operation(self.transform_operation(ops))
            transformed.append(remainded)

        # When transform from conventional to primitive cell, the number of unique operations decreases.
        size = int(np.around(1 / abs(np.linalg.det(self.linear))))
        if size > 1:
            identity = MagneticOperation.identity()
            transformed_set = set(transformed)
            transformed_set.remove(identity)
            transformed = [identity] + list(transformed_set)
            assert len(transformed) * size == len(coset)

        return transformed

    def inverse(self):
        """
        (P, p)^{-1} = (P^{-1}, -P^{-1}p)
        """
        return Transformation(linear=self._invaug[:self.ndim, :self.ndim], origin_shift=self._invaug[:self.ndim, -1])

    @classmethod
    def to_standard(cls, hall_number: int, choice: str, number: int) -> Transformation:
        """
        Transformation to ITA standard setting
        """
        if len(choice) >= 1 and (choice[0] in ["1", "2"]):
            origin_choice = choice[0]
            axes_choice = choice[1:]
        else:
            origin_choice = ""
            axes_choice = choice[:]

        crystal_system = get_crystal_system(hall_number)
        if crystal_system == "monoclinic":
            t_linear = cls._get_monoclinic_transformation_matrix(axes_choice)
        elif crystal_system == "orthorhombic":
            t_linear = cls._get_orthorhombic_transformation_matrix(axes_choice)
        elif crystal_system == "trigonal":
            t_linear = cls._get_trigonal_transformation_matrix(choice)
        else:
            t_linear = np.eye(3)

        # translate inversion center
        origin_shift = None

        if origin_choice == "1":
            # origin shift from 1 to 2
            shift_mapping = {
                48: np.array([1 / 4, 1 / 4, 1 / 4]),
                50: np.array([1 / 4, 1 / 4, 0]),
                59: np.array([1 / 4, 1 / 4, 0]),
                68: np.array([0, 1 / 4, 1 / 4]),
                70: np.array([1 / 8, 1 / 8, 1 / 8]),
                85: np.array([-1 / 4, 1 / 4, 0]),
                86: np.array([-1 / 4, -1 / 4, -1 / 4]),
                88: np.array([0, 1 / 4, 1 / 8]),
                125: np.array([-1 / 4, -1 / 4, 0]),
                126: np.array([-1 / 4, -1 / 4, -1 / 4]),
                129: np.array([-1 / 4, 1 / 4, 0]),
                130: np.array([-1 / 4, 1 / 4, 0]),
                133: np.array([-1 / 4, 1 / 4, -1 / 4]),
                134: np.array([-1 / 4, 1 / 4, -1 / 4]),
                137: np.array([-1 / 4, 1 / 4, -1 / 4]),
                138: np.array([-1 / 4, 1 / 4, -1 / 4]),
                141: np.array([1 / 2, 1 / 4, 1 / 8]),
                142: np.array([1 / 2, 1 / 4, 1 / 8]),
                201: np.array([1 / 4, 1 / 4, 1 / 4]),
                203: np.array([1 / 8, 1 / 8, 1 / 8]),
                222: np.array([1 / 4, 1 / 4, 1 / 4]),
                224: np.array([1 / 4, 1 / 4, 1 / 4]),
                227: np.array([1 / 8, 1 / 8, 1 / 8]),
                228: np.array([3 / 8, 3 / 8, 3 / 8]),
            }
            origin_shift = np.dot(t_linear, shift_mapping[number])

        return Transformation(t_linear, origin_shift)

    @staticmethod
    def _get_monoclinic_transformation_matrix(choice):
        """
        transformation matrix to b1 (ITA standard setting)
        """
        cell_choice = 1
        setting = "b"

        # choice in hall symbols dataset is in '-?[abc][1-3]?'
        if choice[0] == "-":
            setting = choice[:2]  # like "-a"
        else:
            setting = choice[0]  # like "a"

        if choice[-1].isdigit():
            cell_choice = int(choice[-1])  # like 3

        P = np.eye(3)
        # to unique axis b
        if setting != "b":
            P = transformation_matrix_monoclinic["{}_to_b".format(setting)]

        # to choice b1
        if cell_choice == 2:
            P = np.dot(P, transformation_matrix_monoclinic["b2_to_b1"])
        elif cell_choice == 3:
            P = np.dot(P, transformation_matrix_monoclinic["b3_to_b1"])
        return P

    @staticmethod
    def _get_orthorhombic_transformation_matrix(choice):
        if choice == "":
            setting = "abc"
        else:
            setting = choice
        # to abc
        P = transformation_matrix_orthorhombic[setting]
        return P

    @staticmethod
    def _get_trigonal_transformation_matrix(choice):
        if choice == "R":
            # Primitive rhombohedral cell to triple hexagonal cell R1, obverse setting
            return np.array(
                [
                    [1, 0, 1],
                    [-1, 1, 1],
                    [0, -1, 1],
                ],
                dtype=int,
            )
        else:
            # 'H' and ''
            return np.eye(3)


def get_crystal_system(hall_number: int) -> str:
    crystal_system_range = {
        "triclinic": [1, 2],
        "monoclinic": [3, 107],
        "orthorhombic": [108, 348],
        "tetragonal": [349, 429],
        "trigonal": [430, 461],
        "hexagonal": [462, 488],
        "cubic": [489, 530],
    }

    crystal_system = None
    for csystem, (lb, ub) in crystal_system_range.items():
        if (lb <= hall_number) and (hall_number <= ub):
            crystal_system = csystem
    if crystal_system is None:
        raise ValueError("Unknown Hall number: {}".format(hall_number))
    return crystal_system


def get_standard_hall_number(number: int) -> int:
    """
    convert ITA number with standard setting to Hall number
    """
    # fmt: off
    mapping = [
        # triclinic
        1,
        2,
        # monoclinic
        3,  # 3:b
        6,  # 4:b
        9,  # 5:b1
        18,  # 6:b
        21,  # 7:b1
        30,  # 8:b1
        39,  # 9:b1
        57,  # 10:b
        60,  # 11:b
        63,  # 12:b1
        72,  # 13:b1
        81,  # 14:b1
        90,  # 15:b1
        # orthorhombic
        108,  # 16
        109,  # 17
        112,  # 18
        115,  # 19
        116,  # 20
        119,  # 21
        122,  # 22
        123,  # 23
        124,  # 24
        125,  # 25
        128,  # 26
        134,  # 27
        137,  # 28
        143,  # 29
        149,  # 30
        155,  # 31
        161,  # 32
        164,  # 33
        170,  # 34
        173,  # 35
        176,  # 36
        182,  # 37
        185,  # 38
        191,  # 39
        197,  # 40
        203,  # 41
        209,  # 42
        212,  # 43
        215,  # 44
        218,  # 45
        221,  # 46
        227,  # 47
        229,  # 48:2
        230,  # 49
        234,  # 50:2
        239,  # 51
        245,  # 52
        251,  # 53
        257,  # 54
        263,  # 55
        266,  # 56
        269,  # 57
        275,  # 58
        279,  # 59:2
        284,  # 60
        290,  # 61
        292,  # 62
        298,  # 63
        304,  # 64
        310,  # 65
        313,  # 66
        316,  # 67
        323,  # 68:2
        334,  # 69
        336,  # 70:2
        337,  # 71
        338,  # 72
        341,  # 73
        343,  # 74
        # tetragonal
        349,  # 75
        350,  # 76
        351,  # 77
        352,  # 78
        353,  # 79
        354,  # 80
        355,  # 81
        356,  # 82
        357,  # 83
        358,  # 84
        360,  # 85:2
        362,  # 86:2
        363,  # 87
        365,  # 88:2
        366,  # 89
        367,  # 90
        368,  # 91
        369,  # 92
        370,  # 93
        371,  # 94
        372,  # 95
        373,  # 96
        374,  # 97
        375,  # 98
        376,  # 99
        377,  # 100
        378,  # 101
        379,  # 102
        380,  # 103
        381,  # 104
        382,  # 105
        383,  # 106
        384,  # 107
        385,  # 108
        386,  # 109
        387,  # 110
        388,  # 111
        389,  # 112
        390,  # 113
        391,  # 114
        392,  # 115
        393,  # 116
        394,  # 117
        395,  # 118
        396,  # 119
        397,  # 120
        398,  # 121
        399,  # 122
        400,  # 123
        401,  # 124
        403,  # 125:2
        405,  # 126:2
        406,  # 127
        407,  # 128
        409,  # 129:2
        411,  # 130:2
        412,  # 131
        413,  # 132
        415,  # 133:2
        417,  # 134:2
        418,  # 135
        419,  # 136
        421,  # 137:2
        423,  # 138:2
        424,  # 139
        425,  # 140
        427,  # 141:2
        429,  # 142:2
        # trigonal
        430,  # 143
        431,  # 144
        432,  # 145
        433,  # 146:H
        435,  # 147
        436,  # 148:H
        438,  # 149
        439,  # 150
        440,  # 151
        441,  # 152
        442,  # 153
        443,  # 154
        444,  # 155:H
        446,
        447,
        448,
        449,
        450,  # 160:H
        452,  # 161:H
        454,
        455,
        456,
        457,
        458,  # 166:H
        460,  # 167:H
        # hexagonal
        462,
        463,
        464,
        465,
        466,
        467,
        468,
        469,
        470,
        471,
        472,
        473,
        474,
        475,
        476,
        477,
        478,
        479,
        480,
        481,
        482,
        483,
        484,
        485,
        486,
        487,
        488,
        # cubic
        489,
        490,
        491,
        492,
        493,
        494,  # 200
        496,  # 201:2
        497,  # 202
        499,  # 203:2
        500,
        501,
        502,
        503,
        504,
        505,
        506,
        507,
        508,
        509,
        510,
        511,
        512,
        513,
        514,
        515,
        516,
        517,  # 221
        519,  # 222:2
        520,  # 223
        522,  # 224:2
        523,  # 225
        524,  # 226
        526,  # 227:2
        528,  # 228:2
        529,
        530,
    ]
    # fmt: on
    return mapping[number - 1]



"""Tests of delaunay_reduce and niggli_reduce."""

from pathlib import Path

import numpy as np
from spglib import delaunay_reduce, niggli_reduce

cwd = Path(__file__).parent


def _get_lattice_parameters(lattice):
    return np.sqrt(np.dot(lattice, lattice.T).diagonal())


def _get_angles(lattice):
    a, b, c = _get_lattice_parameters(lattice)
    alpha = np.arccos(np.vdot(lattice[1, :], lattice[2, :]) / b / c)
    beta = np.arccos(np.vdot(lattice[2, :], lattice[0, :]) / c / a)
    gamma = np.arccos(np.vdot(lattice[0, :], lattice[1, :]) / a / b)
    return np.array([alpha, beta, gamma]) / np.pi * 180


def _reduce(input_lattices, func, ref_data):
    for i, (input_lattice, reference_lattice) in enumerate(
        zip(input_lattices, ref_data),
    ):
        reduced_lattice = func(input_lattice)
        # self._show_lattice(i, reduced_lattice)
        assert np.allclose(
            [np.linalg.det(input_lattice)],
            [np.linalg.det(reduced_lattice)],
        )
        T = np.dot(np.linalg.inv(reference_lattice.T), input_lattice.T)
        assert np.allclose(T, np.rint(T))
        assert np.allclose(
            _metric_tensor(reduced_lattice),
            _metric_tensor(reference_lattice),
        ), "\n".join(
            [
                "# %d" % (i + 1),
                "Input lattice",
                "%s" % input_lattice,
                " angles: %s" % np.array(_get_angles(input_lattice)),
                "Reduced lattice in reference data",
                "%s" % reference_lattice,
                " angles: %s" % np.array(_get_angles(reference_lattice)),
                "Reduced lattice",
                "%s" % reduced_lattice,
                " angles: %s" % np.array(_get_angles(reduced_lattice)),
                _str_lattice(reduced_lattice),
            ],
        )


def _str_lattice(lattice):
    lines = []
    for v in lattice:
        lines.append(" ".join(["%20.16f" % x for x in v]))
    return "\n".join(lines)


def _metric_tensor(lattice):
    """Return metric tensor.

    Parameters
    ----------
    lattice: Lattice parameters in the form of
        [[a_x, a_y, a_z],
         [b_x, b_y, b_z],
         [c_x, c_y, c_z]]

    Returns
    -------
    Metric tensor:
        [[a.a, a.b, a.c],
         [b.a, b.b, b.c],
         [c.a, c.b, c.c]]

    """
    return np.dot(lattice, np.transpose(lattice))


def _read_file(filename):
    all_lattices = []
    with open(filename) as f:
        lattice = []
        for line in f:
            if line[0] == "#":
                continue
            lattice.append([float(x) for x in line.split()])
            if len(lattice) == 3:
                all_lattices.append(lattice)
                lattice = []
    return np.array(all_lattices)


def _test_niggli_angles(niggli_lattices):
    for i, reference_lattice in enumerate(niggli_lattices):
        angles = np.array(_get_angles(reference_lattice))
        assert (angles > 90 - 1e-3).all() or (angles < 90 + 1e-3).all(), "%d %s" % (
            i + 1,
            angles,
        )


def _show_lattice(self, i, lattice):
    print("# %d" % (i + 1))
    print(self._str_lattice(lattice))
    # for v in lattice:
    #     print(" ".join(["%20.16f" % x for x in v]))


def test_niggli_reduce():
    """Test niggli_reduce."""
    input_lattices = _read_file(cwd / "lattices.dat")
    niggli_lattices = _read_file(cwd / "niggli_lattices.dat")
    _test_niggli_angles(niggli_lattices)
    _reduce(input_lattices, niggli_reduce, niggli_lattices)


def test_delaunay_reduce():
    """Test niggli_reduce."""
    input_lattices = _read_file(cwd / "lattices.dat")
    delaunay_lattices = _read_file(cwd / "delaunay_lattices.dat")
    _reduce(input_lattices, delaunay_reduce, delaunay_lattices)



from __future__ import annotations

from typing import List
from fractions import Fraction
from queue import Queue

import numpy as np
from numpy.typing import NDArray


# Maximal denominator of fractional coordinates of translations
DENOMINATOR = 12


class MagneticOperation:

    ndim = 3

    def __init__(self, augmented_matrix: NDArray, time_reversal: bool):
        self._augmented_matrix = augmented_matrix
        self._time_reversal = time_reversal

        self._linear = self.augmented_matrix[:self.ndim, :self.ndim]
        self._translation = self.augmented_matrix[:self.ndim, -1]

        # For comparison, store linear and translations as tuple of integers
        self._linear_tuple = ndarray_to_integer_tuple(
            self.augmented_matrix[: self.ndim, : self.ndim]
        )
        self._translation_tuple = tuple(
            Fraction(t).limit_denominator(DENOMINATOR)
            for t in self.augmented_matrix[: self.ndim, -1]
        )

    def __eq__(self, other) -> bool:
        if not isinstance(other, MagneticOperation):
            return False

        return (self._linear_tuple == other._linear_tuple) and (
            self._translation_tuple == other._translation_tuple
        ) and (
            self.time_reversal == other.time_reversal
        )

    def __hash__(self):
        return hash((self._linear_tuple, self._translation_tuple, self.time_reversal))

    def __mul__(self, rhs: MagneticOperation) -> MagneticOperation:
        if not isinstance(rhs, MagneticOperation):
            raise ValueError("undefined operation")

        return MagneticOperation(
            augmented_matrix=np.dot(self.augmented_matrix, rhs.augmented_matrix),
            time_reversal=(self.time_reversal != rhs.time_reversal),
        )

    def inverse(self) -> MagneticOperation:
        return MagneticOperation(
            augmented_matrix=np.linalg.inv(self.augmented_matrix),
            time_reversal=self.time_reversal,
        )

    @property
    def linear(self) -> NDArray:
        return self._linear

    @property
    def translation(self) -> NDArray:
        return self._translation

    @property
    def time_reversal(self) -> bool:
        return self._time_reversal

    @property
    def augmented_matrix(self) -> NDArray:
        return self._augmented_matrix

    @classmethod
    def identity(cls) -> MagneticOperation:
        return MagneticOperation.from_linear_translation_time_reversal()

    @classmethod
    def from_linear_translation_time_reversal(cls, linear=None, translation=None, time_reversal=False):
        aug = np.zeros((cls.ndim + 1, cls.ndim + 1))
        aug[-1, -1] = 1

        if linear is None:
            aug[:cls.ndim, :cls.ndim] = np.eye(cls.ndim)
        else:
            aug[:cls.ndim, :cls.ndim] = linear

        if translation is not None:
            aug[:cls.ndim, -1] = translation

        return cls(aug, time_reversal)


def ndarray_to_integer_tuple(array: NDArray):
    array_int = np.around(array).astype(int)
    array_t = tuple(map(tuple, array_int))
    return array_t


def remainder1_with_denominator(arr: NDArray, denominator: int) -> NDArray:
    """
    return arr (mod 1)
    """
    arr_int = np.around(arr * denominator).astype(int)
    arr_int_mod = np.remainder(arr_int, denominator)
    arr_mod = arr_int_mod / denominator
    return arr_mod


def remainder1_symmetry_operation(
    ops: MagneticOperation,
) -> MagneticOperation:
    new_translation = remainder1_with_denominator(ops.translation, DENOMINATOR)
    return MagneticOperation.from_linear_translation_time_reversal(
        ops.linear, new_translation, ops.time_reversal
    )


def traverse(
    generators: List[MagneticOperation],
) -> List[MagneticOperation]:
    """
    Generate all coset operations
    """
    coset = set()
    que = Queue()  # type: ignore
    identity = MagneticOperation.identity()
    que.put(identity)

    while not que.empty():
        g = que.get()
        if g in coset:
            continue
        coset.add(g)

        for h in generators:
            # Take modulus by translation subgroup
            gh = remainder1_symmetry_operation(g * h)
            que.put(gh)

    # Put identity in the first
    coset.remove(identity)
    ret = [identity] + list(coset)  # type: ignore

    return ret
